;; $Id: xtest-harness.scm,v 1.2 2002/05/03 12:56:25 kl Exp kl $

;==============================================================================
; Auxiliary

; A counterpart to list= from SRFI optimized for two lists
(define (xtest-list= = list-a list-b)
  (let rpt ((a list-a) (b list-b))
    (if (null? a)
      (null? b)
      (and (not (null? b))
	   (= (car a) (car b))
	   (rpt (cdr a) (cdr b))))))

; For use in expected results: will be considered as equal to any lambda
(define x-lambda-placeholder (lambda() #t))

; Like equal? but yields #t for comparison of any two procedures.
(define (xtest-equal? x y)
  (cond
    ((and (procedure? x)
	  (procedure? y)))
    ((not (and
	    (list? x)
	    (list? y)))
     (equal? x y))
    (else 
      (xtest-list=
	(lambda (x y) 
	  (xtest-equal? x y))
	x y)
      )))


;==============================================================================
; Test assertions verificators

(define xtest-sep-line "  ==============  ")

; Test assertion: (equal? `(,selector ,@params) expected-result) 
(define-macro (xtest-assert expected-result selector . params)
  (let ((res (gensym)))
    `(begin
       (cerr nl xtest-sep-line ',selector xtest-sep-line  
	     nl "(" ',selector nl 
	     (lambda(port) 
	       (for-each (lambda(p)
			   (xtest-ppw p port))
			 (list ,@params)))
	     ")")
       (let ((,res (,selector ,@params)))
	   (cerr nl ";---->" nl
		 (lambda (port) (pp ,res port)))
	 (or (xtest-equal? ,res ,expected-result)
	   (begin
	     (cerr 
		 "### Expected:" nl
		 (lambda (port) (pp ,expected-result port))
		 "### ASSERTION FAILED for: " ',selector nl)
	     (exit -1)))))))


	
; Test assertion: 
; (begin
;   `(,selector ,@params) 
;   (equal? var expected-result)) 
(define-macro (xtest-assert-var var expected-result selector . params)
  (let ((res (gensym)))
    `(begin
       (let ((xtest--var ,var))
       (cerr nl xtest-sep-line ',selector xtest-sep-line
	     nl "(" ',selector nl 
	     (lambda(port) 
	       (for-each (lambda(p)
			   (xtest-ppw p port))
			 (list ,@params)))
	      ")")
         (,selector ,@params)
	 (cerr nl ";--[var]-->" nl
		 (lambda (port) (pp xtest--var port)))
	 (or (xtest-equal? ,expected-result xtest--var)
	   (begin
	     (cerr  
		 "### Expected:" nl
		 (lambda (port) (pp ,expected-result port))
		 "### ASSERTION FAILED for: " ',selector nl) 
	     (exit -1)))))))

; Test assertion: (equal? `(,selector ,@params) expected-result) 
(define-macro (xtest-assert-write expected-result selector . params)
  (let ((res (gensym)))
    `(begin
       (cerr nl xtest-sep-line ',selector xtest-sep-line  
	     nl "(" ',selector nl 
	     (lambda(port) 
	       (for-each (lambda(p)
			   (xtest-ppw p port))
			 (list ,@params)))
	     ")")
       (let ((,res (,selector ,@params)))
	   (cerr nl ";---->" nl
		 (lambda (port) (write ,res port)))
	 (or (xtest-equal? ,res ,expected-result)
	   (begin
	     (cerr 
		 "### Expected:" nl
		 (lambda (port) (write ,expected-result port))
		 "### ASSERTION FAILED for: " ',selector nl)
	     (exit -1)))))))
